// Copyright (c) 2023 Dominic Masters
// 
// This software is released under the MIT License.
// https://opensource.org/licenses/MIT

#include "TrueTypeTexture.hpp"
#include "assert/assert.hpp"
#include "util/Math.hpp"

using namespace Dawn;

TrueTypeTexture::TrueTypeTexture(
  FT_Face face,
  uint32_t fontSize,
  uint8_t style
)  :
  face(face),
  fontSize(fontSize),
  style(style)
{
  assertTrue(fontSize < 256, "Font size cannot be greater than 256");

  texture = std::make_shared<Texture>();

  // Set freetype font size prior to baking.
  if(FT_Set_Pixel_Sizes(face, 0, fontSize)) {
    assertUnreachable("Failed to set font size");
  }
  
  // Set the texture size
  texture->setSize(
    4096,
    4096,
    TextureFormat::R,
    TextureDataFormat::UNSIGNED_BYTE
  );

  // Texture buffer
  uint8_t *buffer = new uint8_t[texture->getWidth() * texture->getHeight()];
  // Fill with zeros
  std::memset(buffer, 0, texture->getWidth() * texture->getHeight());

  size_t offset = 0;
  struct TrueTypeCharacter info;
  int32_t textureX = 0, textureY = 0;
  int32_t rowHeight = 0;

  // Character sets
  std::vector<wchar_t> characterBlocks;
  // Latin
  for(wchar_t c = 0x0020; c < 0x007F; c++) characterBlocks.push_back(c);
  // Latin-1 Supplement
  for(wchar_t c = 0x00A0; c < 0x00FF; c++) characterBlocks.push_back(c);
  // Latin Extended-A
  for(wchar_t c = 0x0100; c < 0x017F; c++) characterBlocks.push_back(c);
  // Latin Extended-B
  for(wchar_t c = 0x0180; c < 0x024F; c++) characterBlocks.push_back(c);
  // Hiragana
  for(wchar_t c = 0x3040; c < 0x309F; c++) characterBlocks.push_back(c);
  // Katakana
  for(wchar_t c = 0x30A0; c < 0x30FF; c++) characterBlocks.push_back(c);

  // For each character in the character set
  for(wchar_t c : characterBlocks) {
    // Load the character
    if(FT_Load_Char(face, c, FT_LOAD_RENDER)) {
      assertUnreachable("Failed to load character (1)");
    }

    // Store the character information
    info.advance.x = (float_t)(face->glyph->advance.x >> 6);
    info.advance.y = (float_t)(face->glyph->advance.y >> 6);
    info.size = glm::vec2(face->glyph->bitmap.width, face->glyph->bitmap.rows);

    // Determine the texture position
    if(textureX + face->glyph->bitmap.width >= texture->getWidth()) {
      textureX = 0;
      textureY += rowHeight + 1;// Tiny gap between rows
      rowHeight = face->glyph->bitmap.rows;
    } else {
      rowHeight = Math::max<int32_t>(rowHeight, face->glyph->bitmap.rows);
    }

    // Set the quad positions
    info.quad = glm::vec4(
      textureX,
      textureY,
      textureX + face->glyph->bitmap.width,
      textureY + face->glyph->bitmap.rows
    ) / glm::vec4(
      texture->getWidth(),
      texture->getHeight(),
      texture->getWidth(),
      texture->getHeight()
    );

    // Store the cached character data.
    this->characterData[c] = info;

    // Determine pixel offset.
    offset = textureX + (textureY * texture->getWidth());
    assertTrue(
      offset + (face->glyph->bitmap.rows * texture->getWidth()) <=
      texture->getWidth() * texture->getHeight(),
      "Font texture buffer overflow will occur."
    );

    // Buffer pixels, we have to do this one row at a time due to the 
    // differences in width between the glyph and the texture.
    const size_t countPerRow = face->glyph->bitmap.width;
    int32_t i = 0;
    while(i != face->glyph->bitmap.rows) {
      std::memcpy(
        buffer + offset + (i * texture->getWidth()),
        face->glyph->bitmap.buffer + (i * countPerRow),
        countPerRow
      );
      i++;
    }

    // Increment textureX
    textureX += face->glyph->bitmap.width + 1;// I add a tiny gap between chars
  }

  this->texture->buffer(buffer);
  delete[] buffer;
}

struct TrueTypeCharacter TrueTypeTexture::getCharacterData(wchar_t c) {
  return this->characterData[c];
}

TrueTypeTexture::~TrueTypeTexture() {
  FT_Done_Face(this->face);
}